﻿/*
 *  "GEDKeeper", the personal genealogical database editor.
 *  Copyright (C) 2009-2023 by Sergey V. Zhdanovskih.
 *
 *  This file is part of "GEDKeeper".
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using BSLib;
using GKCore.Interfaces;

namespace GKCore
{
    public enum LSID
    {
        /* 000 */ LSID_None,
        /* 001 */ LSID_First,

        /* 001 */ LSID_MIFile = 1,
        /* 002 */ LSID_MIEdit = 2,
        /* 003 */ LSID_MIPedigree,
        /* 004 */ LSID_MIService,
        /* 005 */ LSID_MIWindow,
        /* 006 */ LSID_MIHelp,
        /* 007 */ LSID_MIFileNew,
        /* 008 */ LSID_MIFileLoad,
        /* 009 */ LSID_MIMRUFiles,
        /* 010 */ LSID_MIFileSave,
        /* 011 */ LSID_MIFileClose,
        /* 012 */ LSID_MIFileProperties,
        /* 013 */ LSID_MIExport,
        /* 014 */ LSID_MIExportToFamilyBook,
        /* 015 */ LSID_MIFileSaveAs,
        /* 016 */ LSID_ExportTable,
        /* 017 */ LSID_MIExit,
        /* 018 */ LSID_Import,
        /* 019 */ LSID_Clear,
        /* 020 */ LSID_MIRecordAdd,
        /* 021 */ LSID_MIRecordEdit,
        /* 022 */ LSID_MIRecordDelete,

        /* 023 */ LSID_Deprecated_023, // deprecated

        /* 024 */ LSID_MITreeAncestors,
        /* 025 */ LSID_MITreeDescendants,
        /* 026 */ LSID_MITreeBoth,
        /* 027 */ LSID_MIPedigree_dAboville,
        /* 028 */ LSID_MIPedigree_Konovalov,
        /* 029 */ LSID_MIMap,
        /* 030 */ LSID_MIStats,
        /* 031 */ LSID_MIPedigreeAscend,

        /* 032 */ LSID_MINamesBook, // deprecated
        /* 033 */ LSID_MICalendar, // deprecated
        /* 034 */ LSID_MITimeLine, // deprecated

        /* 035 */ LSID_MIOrganizer,
        /* 036 */ LSID_MIScripts,

        /* 037 */ LSID_Deprecated_037, // deprecated

        /* 038 */ LSID_MITreeTools = 38,
        /* 039 */ LSID_MIFilter,
        /* 040 */ LSID_MIOptions,
        /* 041 */ LSID_MIWinCascade,
        /* 042 */ LSID_MIWinHTile,
        /* 043 */ LSID_MIWinVTile,
        /* 044 */ LSID_MIWinMinimize,
        /* 045 */ LSID_MIWinArrange,

        /* 046 */ LSID_MIGenResources, // deprecated
        /* 047 */ LSID_MIKinshipTerms, // deprecated
        /* 048 */ LSID_MIFAQ, // deprecated

        /* 049 */ LSID_MIContext,
        /* 050 */ LSID_MIAbout,
        /* 051 */ LSID_SBRecords,
        /* 052 */ LSID_SBFiltered,

        /* 053 */ LSID_RPIndividuals,
        /* 054 */ LSID_RPFamilies,
        /* 055 */ LSID_RPNotes,
        /* 056 */ LSID_RPMultimedia,
        /* 057 */ LSID_RPSources,
        /* 058 */ LSID_RPRepositories,
        /* 059 */ LSID_RPGroups,
        /* 060 */ LSID_RPResearches,
        /* 061 */ LSID_RPTasks,
        /* 062 */ LSID_RPCommunications,
        /* 063 */ LSID_RPLocations,

        /* 064 */ LSID_UnkFemale,
        /* 065 */ LSID_UnkMale,
        /* 066 */ LSID_SexU,
        /* 067 */ LSID_SexM,
        /* 068 */ LSID_SexF,
        /* 069 */ LSID_SexX,
        /* 070 */ LSID_FileSaveQuery,
        /* 071 */ LSID_ParentsQuery,
        /* 072 */ LSID_PersonDeleteQuery,
        /* 073 */ LSID_FamilyDeleteQuery,
        /* 074 */ LSID_NoteDeleteQuery,
        /* 075 */ LSID_SourceDeleteQuery,
        /* 076 */ LSID_MediaDeleteQuery,
        /* 077 */ LSID_RepositoryDeleteQuery,
        /* 078 */ LSID_GroupDeleteQuery,
        /* 079 */ LSID_ResearchDeleteQuery,
        /* 080 */ LSID_TaskDeleteQuery,
        /* 081 */ LSID_CommunicationDeleteQuery,
        /* 082 */ LSID_LocationDeleteQuery,
        /* 083 */ LSID_Address,
        /* 084 */ LSID_Events,
        /* 085 */ LSID_Surname,
        /* 086 */ LSID_Name,
        /* 087 */ LSID_Patronymic,
        /* 088 */ LSID_Sex,
        /* 089 */ LSID_Nickname,
        /* 090 */ LSID_SurnamePrefix,
        /* 091 */ LSID_NamePrefix,
        /* 092 */ LSID_NameSuffix,
        /* 093 */ LSID_Patriarch,
        /* 094 */ LSID_Bookmark,
        /* 095 */ LSID_Association,
        /* 096 */ LSID_Relation,
        /* 097 */ LSID_Person,

        /* 098 */ LSID_DlgAccept,
        /* 099 */ LSID_DlgCancel,
        /* 100 */ LSID_DlgClose,

        /* 101 */ LSID_DlgSelect,
        /* 102 */ LSID_DlgAppend,
        /* 103 */ LSID_WinPersonNew,
        /* 104 */ LSID_WinPersonEdit,
        /* 105 */ LSID_WinCheckSex,
        /* 106 */ LSID_WinRecordSelect,
        /* 107 */ LSID_WinSourceCitEdit,
        /* 108 */ LSID_WinUserRefEdit,
        /* 109 */ LSID_Note,
        /* 110 */ LSID_Source,
        /* 111 */ LSID_Page,
        /* 112 */ LSID_Certainty,
        /* 113 */ LSID_Reference,
        /* 114 */ LSID_Type,
        /* 115 */ LSID_Family,
        /* 116 */ LSID_Husband,
        /* 117 */ LSID_Wife,
        /* 118 */ LSID_Status,
        /* 119 */ LSID_Childs,
        /* 120 */ LSID_DetachHusbandQuery,
        /* 121 */ LSID_DetachWifeQuery,
        /* 122 */ LSID_DetachChildQuery,
        /* 123 */ LSID_BirthDate,
        /* 124 */ LSID_DeathDate,
        /* 125 */ LSID_Restriction,
        /* 126 */ LSID_Title,
        /* 127 */ LSID_Members,
        /* 128 */ LSID_WinGroupEdit,
        /* 129 */ LSID_DetachMemberQuery,
        /* 130 */ LSID_MediaView,
        /* 131 */ LSID_Stage,
        /* 132 */ LSID_Telephone,
        /* 133 */ LSID_Mail,
        /* 134 */ LSID_WebSite,
        /* 135 */ LSID_Repository,
        /* 136 */ LSID_Progress,
        /* 137 */ LSID_TimePassed,
        /* 138 */ LSID_TimeRemain,
        /* 139 */ LSID_TimeTotal,
        /* 140 */ LSID_Date,
        /* 141 */ LSID_Text,
        /* 142 */ LSID_ShortTitle,
        /* 143 */ LSID_Author,
        /* 144 */ LSID_Publication,
        /* 145 */ LSID_Common,
        /* 146 */ LSID_DetachRepositoryQuery,
        /* 147 */ LSID_StoreType,
        /* 148 */ LSID_File,
        /* 149 */ LSID_View,
        /* 150 */ LSID_LogHint,
        /* 151 */ LSID_Father,
        /* 152 */ LSID_Mother,
        /* 153 */ LSID_Parents,
        /* 154 */ LSID_Spouses,
        /* 155 */ LSID_Associations,
        /* 156 */ LSID_UserRefs,
        /* 157 */ LSID_Cal_Gregorian,
        /* 158 */ LSID_Cal_Julian,
        /* 159 */ LSID_Cal_Hebrew,
        /* 160 */ LSID_Cal_Islamic,
        /* 161 */ LSID_Cal_Persian,
        /* 162 */ LSID_Cal_Indian,
        /* 163 */ LSID_Cal_Bahai,
        /* 164 */ LSID_Cal_French,
        /* 165 */ LSID_Cal_Roman,
        /* 166 */ LSID_Unknown,
        /* 167 */ LSID_Summary,
        /* 168 */ LSID_Other,
        /* 169 */ LSID_Submission,
        /* 170 */ LSID_Submitter,
        /* 171 */ LSID_Location,
        /* 172 */ LSID_Latitude,
        /* 173 */ LSID_Longitude,
        /* 174 */ LSID_Show,
        /* 175 */ LSID_SearchCoords,
        /* 176 */ LSID_Search,
        /* 177 */ LSID_SelectCoords,
        /* 178 */ LSID_SelectName,
        /* 179 */ LSID_Priority,
        /* 180 */ LSID_Percent,
        /* 181 */ LSID_StartDate,
        /* 182 */ LSID_StopDate,
        /* 183 */ LSID_Goal,
        /* 184 */ LSID_Theme,
        /* 185 */ LSID_Corresponder,
        /* 186 */ LSID_Group,
        /* 187 */ LSID_DetachTaskQuery,
        /* 188 */ LSID_DetachCommunicationQuery,
        /* 189 */ LSID_DetachGroupQuery,
        /* 190 */ LSID_WinResearchEdit,
        /* 191 */ LSID_WinTaskEdit,
        /* 192 */ LSID_WinCommunicationEdit,
        /* 193 */ LSID_Error,
        /* 194 */ LSID_LuaStartFailed,
        /* 195 */ LSID_DateFormatInvalid,
        /* 196 */ LSID_AdCountry,
        /* 197 */ LSID_AdState,
        /* 198 */ LSID_AdCity,
        /* 199 */ LSID_AdPostalCode,
        /* 200 */ LSID_Telephones,

        /* 201 */ LSID_EMails,
        /* 202 */ LSID_WebSites,
        /* 203 */ LSID_Value,
        /* 204 */ LSID_Event,
        /* 205 */ LSID_Place,
        /* 206 */ LSID_Cause,
        /* 207 */ LSID_Agency,
        /* 208 */ LSID_PatFemale,
        /* 209 */ LSID_PatMale,
        /* 210 */ LSID_NotSelectedPerson,
        /* 211 */ LSID_IsNotDefinedSex,
        /* 212 */ LSID_IsNotFamilies, /* not used */
        /* 213 */ LSID_AncestorsNumberIsInvalid,
        /* 214 */ LSID_DescendantsNumberIsInvalid,
        /* 215 */ LSID_GenerationsVisible,
        /* 216 */ LSID_Unlimited,
        /* 217 */ LSID_Spouse,
        /* 218 */ LSID_MarriageDate,
        /* 219 */ LSID_DetachFatherQuery,
        /* 220 */ LSID_DetachMotherQuery,
        /* 221 */ LSID_DetachSpouseQuery,
        /* 222 */ LSID_DetachParentsQuery,
        /* 223 */ LSID_TM_Both,
        /* 224 */ LSID_TM_Ancestors,
        /* 225 */ LSID_TM_Descendants,
        /* 226 */ LSID_TM_TraceSelected,
        /* 227 */ LSID_DoEdit,
        /* 228 */ LSID_FamilyAdd,
        /* 229 */ LSID_SpouseAdd,
        /* 230 */ LSID_SonAdd,
        /* 231 */ LSID_DaughterAdd,
        /* 232 */ LSID_DoDelete,
        /* 233 */ LSID_RebuildTree,
        /* 234 */ LSID_RebuildKinships,
        /* 235 */ LSID_Links,
        /* 236 */ LSID_PlaceAndAttribute,
        /* 237 */ LSID_LMarriage,
        /* 238 */ LSID_LFamily,
        /* 239 */ LSID_Namesakes,
        /* 240 */ LSID_RemoveEventQuery,
        /* 241 */ LSID_DetachNoteQuery,
        /* 242 */ LSID_DetachMultimediaQuery,
        /* 243 */ LSID_DetachSourceQuery,
        /* 244 */ LSID_RemoveAssociationQuery,
        /* 245 */ LSID_RemoveUserRefQuery,
        /* 246 */ LSID_LoadGedComFailed,
        /* 247 */ LSID_CheckGedComFailed,
        /* 248 */ LSID_BirthDays,
        /* 249 */ LSID_DaysRemained,
        /* 250 */ LSID_Interface,
        /* 251 */ LSID_Trees,
        /* 252 */ LSID_Pedigrees,
        /* 253 */ LSID_SaveCoding,
        /* 254 */ LSID_Addresses,
        /* 255 */ LSID_Mails,
        /* 256 */ LSID_Webs,
        /* 257 */ LSID_Internet,
        /* 258 */ LSID_ProxyUse,
        /* 259 */ LSID_ProxyServer,
        /* 260 */ LSID_ProxyPort,
        /* 261 */ LSID_ProxyLogin,
        /* 262 */ LSID_Password,
        /* 263 */ LSID_PasswordIsNotSpecified,
        /* 264 */ LSID_StartupTips,
        /* 265 */ LSID_Language,
        /* 266 */ LSID_ListsAll,
        /* 267 */ LSID_ListPersons,
        /* 268 */ LSID_NamesFormat,
        /* 269 */ LSID_NF1,
        /* 270 */ LSID_NF2,
        /* 271 */ LSID_NF3,
        /* 272 */ LSID_DateFormat,
        /* 273 */ LSID_PlacesWithAddress,
        /* 274 */ LSID_HighlightUnparented,
        /* 275 */ LSID_HighlightUnmarried,
        /* 276 */ LSID_DefList,
        /* 277 */ LSID_ViewTree,
        /* 278 */ LSID_DiffLines,
        /* 279 */ LSID_OnlyYears,
        /* 280 */ LSID_Kinship,
        /* 281 */ LSID_SignsVisible,
        /* 282 */ LSID_TreeDecorative,
        /* 283 */ LSID_PortraitsVisible,
        /* 284 */ LSID_ChildlessExclude,
        /* 285 */ LSID_Decor,
        /* 286 */ LSID_Man,
        /* 287 */ LSID_Woman,
        /* 288 */ LSID_UnkSex,
        /* 289 */ LSID_UnHusband,
        /* 290 */ LSID_UnWife,
        /* 291 */ LSID_Font,
        /* 292 */ LSID_PedigreeGen,
        /* 293 */ LSID_IncludeAttributes,
        /* 294 */ LSID_IncludeNotes,
        /* 295 */ LSID_IncludeSources,
        /* 296 */ LSID_PedigreeFormat,
        /* 297 */ LSID_PF1,
        /* 298 */ LSID_PF2,
        /* 299 */ LSID_RecordGoto,
        /* 300 */ LSID_RecordMoveUp,

        /* 301 */ LSID_RecordMoveDown,
        /* 302 */ LSID_FullName,
        /* 303 */ LSID_BirthPlace,
        /* 304 */ LSID_DeathPlace,
        /* 305 */ LSID_Residence,
        /* 306 */ LSID_Age,
        /* 307 */ LSID_LifeExpectancy,
        /* 308 */ LSID_DaysForBirth,
        /* 309 */ LSID_Religion,
        /* 310 */ LSID_Nationality,
        /* 311 */ LSID_Education,
        /* 312 */ LSID_Occupation,
        /* 313 */ LSID_Caste,
        /* 314 */ LSID_Mili,
        /* 315 */ LSID_MiliInd,
        /* 316 */ LSID_MiliDis,
        /* 317 */ LSID_MiliRank,
        /* 318 */ LSID_Changed,
        /* 319 */ LSID_MarrRegistered,
        /* 320 */ LSID_MarrNotRegistered,
        /* 321 */ LSID_MarrDivorced,
        /* 322 */ LSID_Birth,
        /* 323 */ LSID_Adoption,
        /* 324 */ LSID_Christening,
        /* 325 */ LSID_Graduation,
        /* 326 */ LSID_Retirement,
        /* 327 */ LSID_Naturalization,
        /* 328 */ LSID_Emigration,
        /* 329 */ LSID_Immigration,
        /* 330 */ LSID_Census,
        /* 331 */ LSID_LastWill,
        /* 332 */ LSID_ProbateOfWill,
        /* 333 */ LSID_Death,
        /* 334 */ LSID_Burial,
        /* 335 */ LSID_Cremation,
        /* 336 */ LSID_Fact,
        /* 337 */ LSID_PhysicalDesc,
        /* 338 */ LSID_NationalIDNumber,
        /* 339 */ LSID_SocialSecurityNumber,
        /* 340 */ LSID_ChildsCount,
        /* 341 */ LSID_MarriagesCount,
        /* 342 */ LSID_PersonalProperty,
        /* 343 */ LSID_NobilityTitle,
        /* 344 */ LSID_Travel,
        /* 345 */ LSID_Hobby,
        /* 346 */ LSID_Award,
        /* 347 */ LSID_RK_Unk,
        /* 348 */ LSID_RK_Father,
        /* 349 */ LSID_RK_Mother,
        /* 350 */ LSID_RK_Husband,
        /* 351 */ LSID_RK_Wife,
        /* 352 */ LSID_RK_Son,
        /* 353 */ LSID_RK_Daughter,
        /* 354 */ LSID_RK_Grandfather,
        /* 355 */ LSID_RK_Grandmother,
        /* 356 */ LSID_RK_Grandson,
        /* 357 */ LSID_RK_Granddaughter,
        /* 358 */ LSID_RK_Brother,
        /* 359 */ LSID_RK_Sister,
        /* 360 */ LSID_RK_SonInLaw,
        /* 361 */ LSID_RK_DaughterInLaw,
        /* 362 */ LSID_RK_HusbandFather,
        /* 363 */ LSID_RK_HusbandMother,
        /* 364 */ LSID_RK_WifeFather,
        /* 365 */ LSID_RK_WifeMother,
        /* 366 */ LSID_RK_Uncle,
        /* 367 */ LSID_RK_Aunt,
        /* 368 */ LSID_RK_Nephew,
        /* 369 */ LSID_RK_Niece,
        /* 370 */ LSID_RK_CousinM,
        /* 371 */ LSID_RK_CousinF,
        /* 372 */ LSID_RK_BrotherInLaw_H,
        /* 373 */ LSID_RK_SisterInLaw_H,
        /* 374 */ LSID_RK_BrotherInLaw_W,
        /* 375 */ LSID_RK_SisterInLaw_W,
        /* 376 */ LSID_RK_GreatPrefix,
        /* 377 */ LSID_RK_06,
        /* 378 */ LSID_RK_07,
        /* 379 */ LSID_RK_08,
        /* 380 */ LSID_RK_09,
        /* 381 */ LSID_TooMuchWidth,
        /* 382 */ LSID_Backward,
        /* 383 */ LSID_Forward,
        /* 384 */ LSID_Prev,
        /* 385 */ LSID_Next,
        /* 386 */ LSID_YouKnowWhat,
        /* 387 */ LSID_LoadingLocations,
        /* 388 */ LSID_NotSelected,
        /* 389 */ LSID_MapSelection,
        /* 390 */ LSID_MapSelOnAll,
        /* 391 */ LSID_MSBirthPlaces,
        /* 392 */ LSID_MSDeathPlaces,
        /* 393 */ LSID_MSResiPlace,
        /* 394 */ LSID_MapSelOnSelected,
        /* 395 */ LSID_SaveImage,

        /* 396 */ LSID_FormatUnsupported, // may be deprecated
        /* 397 */ LSID_DataLoadError, // may be deprecated

        /* 398 */ LSID_DocPrint,
        /* 399 */ LSID_DocPreview,
        /* 400 */ LSID_Generation,

        /* 401 */ LSID_CondContains,
        /* 402 */ LSID_CondNotContains,
        /* 403 */ LSID_DK_0,
        /* 404 */ LSID_DK_1,
        /* 405 */ LSID_DK_2,
        /* 406 */ LSID_DK_3,
        /* 407 */ LSID_DK_4,
        /* 408 */ LSID_DK_5,
        /* 409 */ LSID_DK_6,
        /* 410 */ LSID_DK_7,
        /* 411 */ LSID_DK_8,
        /* 412 */ LSID_DK_9,
        /* 413 */ LSID_FEvt_1,
        /* 414 */ LSID_FEvt_2,
        /* 415 */ LSID_FEvt_3,
        /* 416 */ LSID_FEvt_4,
        /* 417 */ LSID_FEvt_5,
        /* 418 */ LSID_FEvt_6,
        /* 419 */ LSID_FEvt_7,
        /* 420 */ LSID_FEvt_8,
        /* 421 */ LSID_FEvt_9,
        /* 422 */ LSID_STRef,
        /* 423 */ LSID_STArc,
        /* 424 */ LSID_STStg,
        /* 425 */ LSID_DefaultValue,
        /* 426 */ LSID_MT_02,
        /* 427 */ LSID_MT_03,
        /* 428 */ LSID_MT_04,
        /* 429 */ LSID_MT_05,
        /* 430 */ LSID_MT_06,
        /* 431 */ LSID_MT_07,
        /* 432 */ LSID_MT_08,
        /* 433 */ LSID_MT_09,
        /* 434 */ LSID_MT_10,
        /* 435 */ LSID_MT_11,
        /* 436 */ LSID_MT_12,
        /* 437 */ LSID_MT_13,
        /* 438 */ LSID_MT_14,
        /* 439 */ LSID_MT_15,
        /* 440 */ LSID_Prt_1,
        /* 441 */ LSID_Prt_2,
        /* 442 */ LSID_Prt_3,
        /* 443 */ LSID_Prt_4,
        /* 444 */ LSID_Prt_5,
        /* 445 */ LSID_RStat_1,
        /* 446 */ LSID_RStat_2,
        /* 447 */ LSID_RStat_3,
        /* 448 */ LSID_RStat_4,
        /* 449 */ LSID_RStat_5,
        /* 450 */ LSID_RStat_6,
        /* 451 */ LSID_Com_1,
        /* 452 */ LSID_Com_2,
        /* 453 */ LSID_Com_3,
        /* 454 */ LSID_Com_4,
        /* 455 */ LSID_Com_5,
        /* 456 */ LSID_Com_6,
        /* 457 */ LSID_CD_1,
        /* 458 */ LSID_CD_2,
        /* 459 */ LSID_G_1,
        /* 460 */ LSID_G_2,
        /* 461 */ LSID_G_3,
        /* 462 */ LSID_G_4,
        /* 463 */ LSID_Cert_1,
        /* 464 */ LSID_Cert_2,
        /* 465 */ LSID_Cert_3,
        /* 466 */ LSID_Cert_4,
        /* 467 */ LSID_Research,
        /* 468 */ LSID_Task,
        /* 469 */ LSID_Communication,
        /* 470 */ LSID_IDsCorrect,
        /* 471 */ LSID_FormatCheck,
        /* 472 */ LSID_EmfFilter,
        /* 473 */ LSID_MainBaseSize,
        /* 474 */ LSID_SyncFin,
        /* 475 */ LSID_PatSearch,
        /* 476 */ LSID_LinksSearch,
        /* 477 */ LSID_MediaFileNotLoaded,
        /* 478 */ LSID_Scale,
        /* 479 */ LSID_GEDCOMFilter,
        /* 480 */ LSID_AllFilter,
        /* 481 */ LSID_ImagesFilter,
        /* 482 */ LSID_ScriptsFilter,
        /* 483 */ LSID_TreeImagesFilter,
        /* 484 */ LSID_DeathCauses,
        /* 485 */ LSID_ExpPedigree,
        /* 486 */ LSID_Parameter,
        /* 487 */ LSID_Total,
        /* 488 */ LSID_ManSum,
        /* 489 */ LSID_WomanSum,
        /* 490 */ LSID_FieldsFilter,

        /* 491 */ LSID_Year,
        /* 492 */ LSID_PDFFilter,
        /* 493 */ LSID_RTFFilter,
        /* 494 */ LSID_HTMLFilter,
        /* 495 */ LSID_FileBackup,
        /* 496 */ LSID_BackupOnlyPrev,
        /* 497 */ LSID_BranchCut,
        /* 498 */ LSID_Not,
        /* 499 */ LSID_BCut_Years,
        /* 500 */ LSID_BCut_Persons,
        /* 501 */ LSID_SrcAll,
        /* 502 */ LSID_SrcNot,
        /* 503 */ LSID_SrcAny,

        /* 504 */ LSID_MasterBase,
        /* 505 */ LSID_OtherBase,
        /* 506 */ LSID_CurrentBase,
        /* 507 */ LSID_IncludedSourceFilter,
        /* 508 */ LSID_IncludedGroupFilter,
        /* 509 */ LSID_DuplicatesSearch,
        /* 510 */ LSID_ValueInvalid,
        /* 511 */ LSID_Operation,
        /* 512 */ LSID_ToolOp_1,
        /* 513 */ LSID_ToolOp_2,
        /* 514 */ LSID_ToolOp_3,
        /* 515 */ LSID_MergeDuplicates,
        /* 516 */ LSID_SearchUnlinkedNamesakes,
        /* 517 */ LSID_ToolOp_6,
        /* 518 */ LSID_ToolOp_7,
        /* 519 */ LSID_ToolOp_8,
        /* 520 */ LSID_ToolOp_9,
        /* 521 */ LSID_SearchMatches,
        /* 522 */ LSID_CheckFamiliesConnection,
        /* 523 */ LSID_All,
        /* 524 */ LSID_OnlyAlive,
        /* 525 */ LSID_OnlyDied,
        /* 526 */ LSID_AliveBefore,
        /* 527 */ LSID_OnlyMans,
        /* 528 */ LSID_OnlyWomans,
        /* 529 */ LSID_NameMask,
        /* 530 */ LSID_PlaceMask,
        /* 531 */ LSID_EventMask,
        /* 532 */ LSID_OnlyPatriarchs,
        /* 533 */ LSID_DateInvalid,
        /* 534 */ LSID_People,
        /* 535 */ LSID_Years,
        /* 536 */ LSID_Decennial,
        /* 537 */ LSID_HowBirthes,
        /* 538 */ LSID_HowDeads,
        /* 539 */ LSID_Living,
        /* 540 */ LSID_Deads,
        /* 541 */ LSID_AvgAge,
        /* 542 */ LSID_AvgLife,
        /* 543 */ LSID_AvgChilds,
        /* 544 */ LSID_AvgBorn,
        /* 545 */ LSID_AvgMarriagesCount,
        /* 546 */ LSID_AvgMarriagesAge,
        /* 547 */ LSID_AncestorsCount,
        /* 548 */ LSID_DescendantsCount,
        /* 549 */ LSID_GenerationsCount,
        /* 550 */ LSID_BirthYears,
        /* 551 */ LSID_BirthYearsDec,
        /* 552 */ LSID_DeathYears,
        /* 553 */ LSID_DeathYearsDec,
        /* 554 */ LSID_DistrChilds,
        /* 555 */ LSID_AgeFirstborn,
        /* 556 */ LSID_MarriagesAge,
        /* 557 */ LSID_DiffSpouses,
        /* 558 */ LSID_SelAll,
        /* 559 */ LSID_SelFamily,
        /* 560 */ LSID_SelAncestors,
        /* 561 */ LSID_SelDescendants,
        /* 562 */ LSID_RecMerge,
        /* 563 */ LSID_RM_Search,
        /* 564 */ LSID_RM_Skip,
        /* 565 */ LSID_RM_Records,
        /* 566 */ LSID_RM_SearchPersons,

        /* 567 */ LSID_RM_DirectMatching, // deprecated
        /* 568 */ LSID_RM_IndistinctMatching,
        /* 569 */ LSID_RM_OnlyNP, // deprecated

        /* 570 */ LSID_RM_BirthYear,
        /* 571 */ LSID_RM_NameAccuracy,
        /* 572 */ LSID_RM_YearInaccuracy,
        /* 573 */ LSID_Repair,
        /* 574 */ LSID_MinGenerations,
        /* 575 */ LSID_SetPatFlag,
        /* 576 */ LSID_InsertIntoBook,
        /* 577 */ LSID_SimilarSurnames,
        /* 578 */ LSID_SimilarNames,
        /* 579 */ LSID_RecsDeleted,
        /* 580 */ LSID_Record,
        /* 581 */ LSID_Problem,
        /* 582 */ LSID_Solve,
        /* 583 */ LSID_PersonLonglived,
        /* 584 */ LSID_PersonSexless,
        /* 585 */ LSID_LiveYearsInvalid,
        /* 586 */ LSID_StrangeSpouse,
        /* 587 */ LSID_StrangeParent,
        /* 588 */ LSID_Descendants,
        /* 589 */ LSID_Generations,
        /* 590 */ LSID_LinksCount,
        /* 591 */ LSID_PlacesPrepare,
        /* 592 */ LSID_PlaceAlreadyInBook,
        /* 593 */ LSID_DNAMarkers,
        /* 594 */ LSID_AAF_1,
        /* 595 */ LSID_AAF_2,
        /* 596 */ LSID_EmptyFamily,
        /* 597 */ LSID_BackupEachRevision,
        /* 598 */ LSID_Loading,
        /* 599 */ LSID_FileSaveError,
        /* 600 */ LSID_FillColor,

        /* 601 */ LSID_FillImage,
        /* 602 */ LSID_TreeIndividualsCount,
        /* 603 */ LSID_LogView,
        /* 604 */ LSID_AncestorsCircle,
        /* 605 */ LSID_Circle,
        /* 606 */ LSID_TextColor,
        /* 607 */ LSID_BackColor,
        /* 608 */ LSID_LinesColor,
        /* 609 */ LSID_RestrictNone,
        /* 610 */ LSID_RestrictLocked,
        /* 611 */ LSID_RestrictConfidential,
        /* 612 */ LSID_RestrictPrivacy,
        /* 613 */ LSID_LogSend,
        /* 614 */ LSID_Charts,
        /* 615 */ LSID_SizeToFit,
        /* 616 */ LSID_ZoomIn,
        /* 617 */ LSID_ZoomOut,
        /* 618 */ LSID_SetIsDead,
        /* 619 */ LSID_DefineSex,
        /* 620 */ LSID_Plugins,
        /* 621 */ LSID_TM_TraceKinships,
        /* 622 */ LSID_NoMatchesFound,
        /* 623 */ LSID_FindPrevious,
        /* 624 */ LSID_FindNext,
        /* 625 */ LSID_CertaintyIndex,
        /* 626 */ LSID_BirthByMonth,
        /* 627 */ LSID_Month,
        /* 628 */ LSID_DlgReset,
        /* 629 */ LSID_Field,
        /* 630 */ LSID_Condition,
        /* 631 */ LSID_DoNotSetSource,
        /* 632 */ LSID_Slideshow,
        /* 633 */ LSID_PrevRec,
        /* 634 */ LSID_NextRec,
        /* 635 */ LSID_GenerationFailed,
        /* 636 */ LSID_Start,
        /* 637 */ LSID_Stop,
        /* 638 */ LSID_Demography,
        /* 639 */ LSID_Names,
        /* 640 */ LSID_NTAka,
        /* 641 */ LSID_NTBirth,
        /* 642 */ LSID_NTImmigrant,
        /* 643 */ LSID_NTMaiden,
        /* 644 */ LSID_NTMarried,
        /* 645 */ LSID_RemoveNameQuery,
        /* 646 */ LSID_RemoveNameFailed,
        /* 647 */ LSID_Quote,
        /* 648 */ LSID_BookmarkMerged,
        /* 649 */ LSID_ShowDatesCalendar,
        /* 650 */ LSID_RecordDuplicate,
        /* 651 */ LSID_DuplicateWarning,
        /* 652 */ LSID_ShowDatesSigns,
        /* 653 */ LSID_FileNewTip,
        /* 654 */ LSID_FileLoadTip,
        /* 655 */ LSID_FileSaveTip,
        /* 656 */ LSID_RecordAddTip,
        /* 657 */ LSID_RecordEditTip,
        /* 658 */ LSID_RecordDeleteTip,
        /* 659 */ LSID_FilterTip,
        /* 660 */ LSID_TreeAncestorsTip,
        /* 661 */ LSID_TreeDescendantsTip,
        /* 662 */ LSID_TreeBothTip,
        /* 663 */ LSID_PedigreeTip,
        /* 664 */ LSID_Pedigree_dAbovilleTip,
        /* 665 */ LSID_Pedigree_KonovalovTip,
        /* 666 */ LSID_StatsTip,
        /* 667 */ LSID_ImageSaveTip,
        /* 668 */ LSID_ModesTip,
        /* 669 */ LSID_NewScriptTip,
        /* 670 */ LSID_LoadScriptTip,
        /* 671 */ LSID_SaveScriptTip,
        /* 672 */ LSID_RunScriptTip,
        /* 673 */ LSID_PortraitAddTip,
        /* 674 */ LSID_PortraitDeleteTip,
        /* 675 */ LSID_ParentsAddTip,
        /* 676 */ LSID_ParentsEditTip,
        /* 677 */ LSID_ParentsDeleteTip,
        /* 678 */ LSID_FatherAddTip,
        /* 679 */ LSID_FatherDeleteTip,
        /* 680 */ LSID_FatherSelTip,
        /* 681 */ LSID_MotherAddTip,
        /* 682 */ LSID_MotherDeleteTip,
        /* 683 */ LSID_MotherSelTip,
        /* 684 */ LSID_NameCopyTip,
        /* 685 */ LSID_HusbandAddTip,
        /* 686 */ LSID_HusbandDeleteTip,
        /* 687 */ LSID_HusbandSelTip,
        /* 688 */ LSID_WifeAddTip,
        /* 689 */ LSID_WifeDeleteTip,
        /* 690 */ LSID_WifeSelTip,
        /* 691 */ LSID_PersonAttachTip,
        /* 692 */ LSID_PlaceAddTip,
        /* 693 */ LSID_PlaceDeleteTip,
        /* 694 */ LSID_ShowOnMapTip,
        /* 695 */ LSID_SourceAddTip,
        /* 696 */ LSID_GoalSelectTip,
        /* 697 */ LSID_NewDBFileNeedToSave,
        /* 698 */ LSID_FileWithSameNameAlreadyExists,
        /* 699 */ LSID_ItsNotGEDSECCompatibleFile,
        /* 700 */ LSID_ThisPersonHasSeveralFamilies,

        /* 701 */ LSID_UploadErrorInExcel,
        /* 702 */ LSID_PersonsFilter,
        /* 703 */ LSID_LinesVisible,
        /* 704 */ LSID_MatchType,
        /* 705 */ LSID_MatchInternal,
        /* 706 */ LSID_MathExternal,
        /* 707 */ LSID_Analyze,
        /* 708 */ LSID_Match,
        /* 709 */ LSID_WithoutDates,
        /* 710 */ LSID_PatriarchsDiagram,
        /* 711 */ LSID_FamilyBook,
        /* 712 */ LSID_TableOfContents,
        /* 713 */ LSID_PersonalRecords,
        /* 714 */ LSID_Indexes,
        /* 715 */ LSID_NotDeterminedPersonSex,
        /* 716 */ LSID_Autosave,
        /* 717 */ LSID_Minutes,
        /* 718 */ LSID_IncludeGenerations,
        /* 719 */ LSID_HideEmptySegments,
        /* 720 */ LSID_DescendantsCircle,
        /* 721 */ LSID_PortraitSelect,
        /* 722 */ LSID_RelationshipCalculator,
        /* 723 */ LSID_RelationshipMask,
        /* 724 */ LSID_BirthdayToday,
        /* 725 */ LSID_BirthdayTomorrow,
        /* 726 */ LSID_BloodGroup,
        /* 727 */ LSID_HairColor,
        /* 728 */ LSID_EyesColor,
        /* 729 */ LSID_MDNAHaplogroup,
        /* 730 */ LSID_YDNAHaplogroup,
        /* 731 */ LSID_RecordIsLocked,
        /* 732 */ LSID_MaidenSurname,
        /* 733 */ LSID_MarriedSurname,
        /* 734 */ LSID_AdditionalNames, // TODO: better text
        /* 735 */ LSID_ExtendedWomanSurnames,
        /* 736 */ LSID_WSF_Maiden_Married,
        /* 737 */ LSID_WSF_Married_Maiden,
        /* 738 */ LSID_WSF_Maiden,
        /* 739 */ LSID_WSF_Married,
        /* 740 */ LSID_HusbSign,
        /* 741 */ LSID_WifeSign,
        /* 742 */ LSID_Geocoder,
        /* 743 */ LSID_FatherAdd,
        /* 744 */ LSID_MotherAdd,
        /* 745 */ LSID_DefaultPortraits,
        /* 746 */ LSID_RemovableMediaWarningMessage,
        /* 747 */ LSID_RemovableMediaWarningOption,
        /* 748 */ LSID_LoadRecentFiles,
        /* 749 */ LSID_HolidayToday,
        /* 750 */ LSID_HolidayTomorrow,
        /* 751 */ LSID_DaysRemainedBeforeHoliday,
        /* 752 */ LSID_Holidays,
        /* 753 */ LSID_EmbeddedMediaPlayer,
        /* 754 */ LSID_AllowMediaDirectReferences,
        /* 755 */ LSID_CopyingFile,
        /* 756 */ LSID_Actions,
        /* 757 */ LSID_SelectAndCopy,
        /* 758 */ LSID_NumberSym,
        /* 759 */ LSID_Reports,
        /* 760 */ LSID_InvertedTree,
        /* 761 */ LSID_MarriagesDates,
        /* 762 */ LSID_AutoCheckUpdates,
        /* 763 */ LSID_ArcText,
        /* 764 */ LSID_TreesAlbum,
        /* 765 */ LSID_ShowPlaces,
        /* 766 */ LSID_HideUnknownSpouses,
        /* 767 */ LSID_Spacings,
        /* 768 */ LSID_Margins,
        /* 769 */ LSID_BranchDist,
        /* 770 */ LSID_GenDist,
        /* 771 */ LSID_SpouseDist,
        /* 772 */ LSID_ImageSize,
        /* 773 */ LSID_DetectedDataLoop,
        /* 774 */ LSID_FatherAsChild,
        /* 775 */ LSID_MotherAsChild,
        /* 776 */ LSID_DuplicateChildrenInFamily,
        /* 777 */ LSID_SelectColor,
        /* 778 */ LSID_AutoSortChildren,
        /* 779 */ LSID_AutoSortSpouses,
        /* 780 */ LSID_CheckTreeSize,
        /* 781 */ LSID_PLT_Adopted,
        /* 782 */ LSID_PLT_Birth,
        /* 783 */ LSID_PLT_Foster,
        /* 784 */ LSID_CharsetDetection,
        /* 785 */ LSID_LinkageType,
        /* 786 */ LSID_HeWasBorn,
        /* 787 */ LSID_SheWasBorn,
        /* 788 */ LSID_GedMLFilter,
        /* 789 */ LSID_ChildWithoutParents,
        /* 790 */ LSID_FamilyRecordWithoutFamily,
        /* 791 */ LSID_MediaFileDeleteQuery,
        /* 792 */ LSID_StgNotFound,
        /* 793 */ LSID_FileNotFound,
        /* 794 */ LSID_AllowMediaRelativeReferences,
        /* 795 */ LSID_MediaStoreDefault,
        /* 796 */ LSID_AllowDeleteMediaFileFromStgArc,
        /* 797 */ LSID_AllowDeleteMediaFileFromRefs,
        /* 798 */ LSID_DeleteMediaFileWithoutConfirm,
        /* 799 */ LSID_MediaRecordWithoutFiles,
        /* 800 */ LSID_ArcNotFound,

        /* 801 */ LSID_RI_GeorgeKnight = 801,
        /* 802 */ LSID_USSR_WWII_Combatant,
        /* 803 */ LSID_USSR_WWII_KilledInBattle,
        /* 804 */ LSID_USSR_WWII_WorkerInRear,
        /* 805 */ LSID_USSR_Repressed,
        /* 806 */ LSID_Religion_Islam,
        /* 807 */ LSID_Religion_Catholicism,
        /* 808 */ LSID_Religion_Orthodoxy,
        /* 809 */ LSID_Religion_TheOldBelievers,

        /* 810 */ LSID_STRel,
        /* 811 */ LSID_BackupRevisionsMaxCount,
        /* 812 */ LSID_FirstCapitalLetterInNames,
        /* 813 */ LSID_DefaultDepth,
        /* 814 */ LSID_WarningOfDialogUnsavedChanges,
        /* 815 */ LSID_WarnForClosingDialog,
        /* 816 */ LSID_DottedLinesOfAdoptedChildren,
        /* 817 */ LSID_GoToPersonRecord,
        /* 818 */ LSID_STWeb,
        /* 819 */ LSID_SeparateDatesAndPlacesLines,
        /* 820 */ LSID_Details,
        /* 821 */ LSID_CopyXRef,
        /* 822 */ LSID_BoldNames,

        /* 823 */ LSID_Baptism,
        /* 824 */ LSID_BarMitzvah,
        /* 825 */ LSID_BatMitzvah,
        /* 826 */ LSID_Blessing,
        /* 827 */ LSID_AdultChristening,
        /* 828 */ LSID_Confirmation,
        /* 829 */ LSID_FirstCommunion,
        /* 830 */ LSID_Ordination,
        /* 831 */ LSID_GeoSearchCountryRestriction,
        /* 832 */ LSID_SeparateDepth,
        /* 833 */ LSID_Ancestors,
        /* 834 */ LSID_MatchesNotFound,
        /* 835 */ LSID_ShortKinshipForm,
        /* 836 */ LSID_SurnameFirstInOrder,
        /* 837 */ LSID_Borders,
        /* 838 */ LSID_UpdateToLatestVersion,
        /* 839 */ LSID_Providers,
        /* 840 */ LSID_Coordinates,
        /* 841 */ LSID_LoadPlaces,
        /* 842 */ LSID_OnlyLocality,
        /* 843 */ LSID_GoToPrimaryBranch,
        /* 844 */ LSID_MinimizingWidth,
        /* 845 */ LSID_ContinueQuestion,
        /* 846 */ LSID_ShowAge,
        /* 847 */ LSID_SurnameInCapitals,
        /* 848 */ LSID_SvgFilter,
        /* 849 */ LSID_CertaintyAlgorithm,
        /* 850 */ LSID_LocalizedCalendarSignaturesOptLabel,
        /* 851 */ LSID_LocalizedCalendarSignaturesValues,
        /* 852 */ LSID_Copy,
        /* 853 */ LSID_Cut,
        /* 854 */ LSID_Paste,
        /* 855 */ LSID_PersonHasHalfLinkOfSpouseToFamily,
        /* 856 */ LSID_PersonHasHalfLinkOfChildToFamily,
        /* 857 */ LSID_FamilyHasHalfLinkOfHusbandToPerson,
        /* 858 */ LSID_FamilyHasHalfLinkOfWifeToPerson,
        /* 859 */ LSID_FamilyHasHalfLinkOfChildToParentsFamily,
        /* 860 */ LSID_GarbledSpouses,
        /* 861 */ LSID_USSR_CPSUMember,
        /* 862 */ LSID_RepoRecsMergeWarning,
        /* 863 */ LSID_UseExtraControls,
        /* 864 */ LSID_SeveralFamiliesOfParents,
        /* 865 */ LSID_Specials,
        /* 866 */ LSID_UseInlineImagesInSvg,
        /* 867 */ LSID_UseExtendedNotes,
        /* 868 */ LSID_KeepRichNames,
        /* 869 */ LSID_MaximizeChartWindows,
        /* 870 */ LSID_ExtendedTree,
        /* 871 */ LSID_FanMode,
        /* 872 */ LSID_OpenInNewWindow,
        /* 873 */ LSID_UnknownPlaceOfPerson,
        /* 874 */ LSID_CheckPersonPlaces,
        /* 875 */ LSID_HighlightInaccessibleFiles,
        /* 876 */ LSID_DifferenceBetweenSpousesIsAbnormallyHigh,
        /* 877 */ LSID_DifferenceBetweenSiblingsIsAbnormallyHigh,

        /* 878 */ LSID_FindAndReplace,
        /* 879 */ LSID_Find,
        /* 880 */ LSID_ReplaceWith,
        /* 881 */ LSID_MatchCase,
        /* 882 */ LSID_MatchWildcards,
        /* 883 */ LSID_WholeWord,
        /* 884 */ LSID_Replace,
        /* 885 */ LSID_ReplaceAll,
        /* 886 */ LSID_Property,

        /* 887 */ LSID_Themes,
        /* 888 */ LSID_URTreeNoteType,
        /* 889 */ LSID_ShowTreeNotes,
        /* 890 */ LSID_NoGoalRecordSpecified,
        /* 891 */ LSID_InvalidFileName,
        /* 892 */ LSID_SearchAndFilterByAllNames,
        /* 893 */ LSID_DataQuality,
        /* 894 */ LSID_DQHint,
        /* 895 */ LSID_Refresh,
        /* 896 */ LSID_ResetFilter,
        /* 897 */ LSID_ShortenDateRanges,
        /* 898 */ LSID_PressEnterToSaveFilter,

        /* 899 */ LSID_AnniversaryToday,
        /* 900 */ LSID_AnniversaryTomorrow,
        /* 901 */ LSID_AnniversaryDaysRemained,

        /* 000 */ LSID_Last = LSID_AnniversaryDaysRemained
    }


    public static class LangMan
    {
        public const int LS_DEF_CODE = 1033;
        public const string LS_DEF_SIGN = "enu";
        public const string LS_DEF_NAME = "English";

        private static readonly string[] LSDefList = new string[]
        {
            /* 001 */ "File",
            /* 002 */ "Edit",
            /* 003 */ "Pedigree",
            /* 004 */ "Service",
            /* 005 */ "&Windows",
            /* 006 */ "Help",
            /* 007 */ "New",
            /* 008 */ "Open...",
            /* 009 */ "Reopen",
            /* 010 */ "Save",
            /* 011 */ "Close",
            /* 012 */ "Properties",
            /* 013 */ "Export",
            /* 014 */ "Book of families...",
            /* 015 */ "Save As...",
            /* 016 */ "Export table...",
            /* 017 */ "Exit",
            /* 018 */ "Import",
            /* 019 */ "Clear",
            /* 020 */ "Add record",
            /* 021 */ "Edit record",
            /* 022 */ "Delete record",
            /* 023 */ "<?>",
            /* 024 */ "Ancestors tree",
            /* 025 */ "Descendants tree",
            /* 026 */ "Complete tree",
            /* 027 */ "d'Aboville descending pedigree",
            /* 028 */ "Konovalov descending pedigree",
            /* 029 */ "Maps",
            /* 030 */ "Statistics",
            /* 031 */ "Ascending pedigree",
            /* 032 */ "<?>",
            /* 033 */ "<?>",
            /* 034 */ "<?>",
            /* 035 */ "Organizer",
            /* 036 */ "Scripts...",
            /* 037 */ "<?>",
            /* 038 */ "Tools...",
            /* 039 */ "Filter",
            /* 040 */ "Options",
            /* 041 */ "&Cascade",
            /* 042 */ "&Horizontal tile",
            /* 043 */ "&Vertical tile",
            /* 044 */ "&Minimize all",
            /* 045 */ "&Rearrange all",
            /* 046 */ "<?>",
            /* 047 */ "<?>",
            /* 048 */ "<?>",
            /* 049 */ "Content",
            /* 050 */ "About",
            /* 051 */ "Records",
            /* 052 */ "filtered",
            /* 053 */ "Individuals",
            /* 054 */ "Families",
            /* 055 */ "Notes",
            /* 056 */ "Multimedia",
            /* 057 */ "Sources",
            /* 058 */ "Repositories",
            /* 059 */ "Groups",
            /* 060 */ "Researches",
            /* 061 */ "Tasks",
            /* 062 */ "Communications",
            /* 063 */ "Locations",
            /* 064 */ "unknown",
            /* 065 */ "unknown",
            /* 066 */ "?",
            /* 067 */ "Male",
            /* 068 */ "Female",
            /* 069 */ "Intersex",
            /* 070 */ "Do you want to save changes to the modified file?",
            /* 071 */ "The specified parent has an existed family \"{0}\". Do you want to add the child to that family?",
            /* 072 */ "Are you sure you want to remove individual \"{0}\"?",
            /* 073 */ "Are you sure you want to remove family \"{0}\"?",
            /* 074 */ "Are you sure you want to remove note \"{0}\"?",
            /* 075 */ "Are you sure you want to remove source \"{0}\"?",
            /* 076 */ "Are you sure you want to remove multimedia \"{0}\"?",
            /* 077 */ "Are you sure you want to remove repository \"{0}\"?",
            /* 078 */ "Are you sure you want to remove group \"{0}\"?",
            /* 079 */ "Are you sure you want to remove research \"{0}\"?",
            /* 080 */ "Are you sure you want to remove task \"{0}\"?",
            /* 081 */ "Are you sure you want to remove communication \"{0}\"?",
            /* 082 */ "Are you sure you want to remove location \"{0}\"?",
            /* 083 */ "Address",
            /* 084 */ "Events",
            /* 085 */ "Surname",
            /* 086 */ "Name",
            /* 087 */ "Patronymic",
            /* 088 */ "Sex",
            /* 089 */ "Nickname",
            /* 090 */ "Last name prefix",
            /* 091 */ "Name prefix",
            /* 092 */ "Name suffix",
            /* 093 */ "Patriarch",
            /* 094 */ "Bookmark",
            /* 095 */ "Association",
            /* 096 */ "Relation",
            /* 097 */ "Individual",
            /* 098 */ "Apply",
            /* 099 */ "Cancel",
            /* 100 */ "Close",
            /* 101 */ "Select",
            /* 102 */ "Add new",
            /* 103 */ "New individual",
            /* 104 */ "Edit individual",
            /* 105 */ "Checking of sex",
            /* 106 */ "Selecting a record",
            /* 107 */ "Source citation",
            /* 108 */ "User reference",
            /* 109 */ "Note",
            /* 110 */ "Source",
            /* 111 */ "Page",
            /* 112 */ "Certainty",
            /* 113 */ "Reference",
            /* 114 */ "Type",
            /* 115 */ "Family",
            /* 116 */ "Husband",
            /* 117 */ "Wife",
            /* 118 */ "Status",
            /* 119 */ "Children",
            /* 120 */ "Are you sure you want to detach link to husband?",
            /* 121 */ "Are you sure you want to detach link to wife?",
            /* 122 */ "Are you sure you want to detach link to child?",
            /* 123 */ "Birth date",
            /* 124 */ "Death date",
            /* 125 */ "Security restriction",
            /* 126 */ "Title",
            /* 127 */ "Members",
            /* 128 */ "Edit group",
            /* 129 */ "Are you sure you want to detach link to member?",
            /* 130 */ "view",
            /* 131 */ "Stage",
            /* 132 */ "Telephone number",
            /* 133 */ "email",
            /* 134 */ "Web site",
            /* 135 */ "Repository",
            /* 136 */ "Progress",
            /* 137 */ "Elapsed time",
            /* 138 */ "Remaining time",
            /* 139 */ "Total time",
            /* 140 */ "Date",
            /* 141 */ "Text",
            /* 142 */ "Short title",
            /* 143 */ "Author",
            /* 144 */ "Published",
            /* 145 */ "Common",
            /* 146 */ "Are you sure you want to detach link to repository?",
            /* 147 */ "Store type",
            /* 148 */ "File",
            /* 149 */ "View",
            /* 150 */ "Fragment {0}: {1} individuals",
            /* 151 */ "Father",
            /* 152 */ "Mother",
            /* 153 */ "Parents",
            /* 154 */ "Spouses",
            /* 155 */ "Associations",
            /* 156 */ "User references",
            /* 157 */ "Gregorian",
            /* 158 */ "Julian",
            /* 159 */ "Hebrew",
            /* 160 */ "Islamic",
            /* 161 */ "Persian",
            /* 162 */ "Indian",
            /* 163 */ "Bahai",
            /* 164 */ "French",
            /* 165 */ "Roman",
            /* 166 */ "Unknown",
            /* 167 */ "Summary",
            /* 168 */ "Other",
            /* 169 */ "Submission",
            /* 170 */ "Submitter",
            /* 171 */ "Location",
            /* 172 */ "Latitude",
            /* 173 */ "Longitude",
            /* 174 */ "Show",
            /* 175 */ "Search coords",
            /* 176 */ "Search",
            /* 177 */ "Select coords",
            /* 178 */ "Select name",
            /* 179 */ "Priority",
            /* 180 */ "Percent",
            /* 181 */ "Started",
            /* 182 */ "Stopped",
            /* 183 */ "Goal",
            /* 184 */ "Theme",
            /* 185 */ "Corresponder",
            /* 186 */ "Group",
            /* 187 */ "Are you sure you want to detach link to task?",
            /* 188 */ "Are you sure you want to detach link to communication?",
            /* 189 */ "Are you sure you want to detach link to group?",
            /* 190 */ "Edit research",
            /* 191 */ "Edit task",
            /* 192 */ "Edit communication",
            /* 193 */ "Error",
            /* 194 */ "Lua start failed!",
            /* 195 */ "Date format is invalid",
            /* 196 */ "Country",
            /* 197 */ "State",
            /* 198 */ "City",
            /* 199 */ "Postal code",
            /* 200 */ "Telephone numbers",
            /* 201 */ "Emails",
            /* 202 */ "Web sites",
            /* 203 */ "Value",
            /* 204 */ "Event",
            /* 205 */ "Place",
            /* 206 */ "Cause",
            /* 207 */ "Agency",
            /* 208 */ "Female",
            /* 209 */ "Male",
            /* 210 */ "An individual record is not selected.",
            /* 211 */ "This individual has no sex defined.",
            /* 212 */ "This individual has no families.",
            /* 213 */ "Estimated number of ancestors {0} is more than permissible limits.",
            /* 214 */ "Estimated number of descendants {0} is more than permissible limits.",
            /* 215 */ "Generations visible:",
            /* 216 */ "unlimited",
            /* 217 */ "Spouse",
            /* 218 */ "Marriage date",
            /* 219 */ "Are you sure you want to detach link to father?",
            /* 220 */ "Are you sure you want to detach link to mother?",
            /* 221 */ "Are you sure you want to detach link to spouse?",
            /* 222 */ "Are you sure you want to detach link to parents?",
            /* 223 */ "All",
            /* 224 */ "Ancestors only",
            /* 225 */ "Descendants only",
            /* 226 */ "Trace root",
            /* 227 */ "Edit",
            /* 228 */ "Add family",
            /* 229 */ "Add spouse",
            /* 230 */ "Add son",
            /* 231 */ "Add daughter",
            /* 232 */ "Delete",
            /* 233 */ "Rebuild tree",
            /* 234 */ "Rebuild kinships",
            /* 235 */ "Links",
            /* 236 */ "Place and attribute",
            /* 237 */ "marriage",
            /* 238 */ "family",
            /* 239 */ "Namesakes",
            /* 240 */ "Remove the event?",
            /* 241 */ "Are you sure you want to detach link to note?",
            /* 242 */ "Are you sure you want to detach link to multimedia?",
            /* 243 */ "Are you sure you want to detach link to source?",
            /* 244 */ "Are you sure you want to remove the association?",
            /* 245 */ "Are you sure you want to remove reference \"{0}\"?",
            /* 246 */ "Loading of gedcom failed.",
            /* 247 */ "Check gedcom format failed.",
            /* 248 */ "Birth days",
            /* 249 */ "{1} days remain before the birthday of \"{0}\"",
            /* 250 */ "Interface",
            /* 251 */ "Trees",
            /* 252 */ "Pedigrees",
            /* 253 */ "Files encoding",
            /* 254 */ "Addresses",
            /* 255 */ "Mails",
            /* 256 */ "Webs",
            /* 257 */ "Internet",
            /* 258 */ "Use proxy",
            /* 259 */ "Server",
            /* 260 */ "Port",
            /* 261 */ "Login",
            /* 262 */ "Password",
            /* 263 */ "The password is not specified",
            /* 264 */ "Show tips on startup",
            /* 265 */ "Language",
            /* 266 */ "All lists",
            /* 267 */ "List of individuals",
            /* 268 */ "Name format in lists",
            /* 269 */ "Surname_Name_Patronymic",
            /* 270 */ "Surname; Name_Patronymic",
            /* 271 */ "Surname; Name; Patronymic",
            /* 272 */ "Date format in lists",
            /* 273 */ "Include address into place text",
            /* 274 */ "Highlight unparented",
            /* 275 */ "Highlight unmarried",
            /* 276 */ "Default values",
            /* 277 */ "Show individuals in tree",
            /* 278 */ "Separate lines",
            /* 279 */ "Only years",
            /* 280 */ "Kinship",
            /* 281 */ "Additional symbols",
            /* 282 */ "Tree decorative",
            /* 283 */ "Show portraits",
            /* 284 */ "Hide died in childhood",
            /* 285 */ "Decor",
            /* 286 */ "Man",
            /* 287 */ "Woman",
            /* 288 */ "Unknown sex",
            /* 289 */ "Divorcee husband",
            /* 290 */ "Divorcee wife",
            /* 291 */ "Font",
            /* 292 */ "Pedigree generation",
            /* 293 */ "Include attributes of individuals",
            /* 294 */ "Include notes",
            /* 295 */ "Include sources",
            /* 296 */ "Pedigree format",
            /* 297 */ "Excess",
            /* 298 */ "Compact",
            /* 299 */ "Jump to a record",
            /* 300 */ "Move record up",
            /* 301 */ "Move record down",
            /* 302 */ "Full name",
            /* 303 */ "Birth place",
            /* 304 */ "Death place",
            /* 305 */ "Residence",
            /* 306 */ "Age",
            /* 307 */ "Lifetime",
            /* 308 */ "Days before birth",
            /* 309 */ "Religion",
            /* 310 */ "Nationality",
            /* 311 */ "Education",
            /* 312 */ "Occupation",
            /* 313 */ "Caste",
            /* 314 */ "Mili",
            /* 315 */ "Mili ind",
            /* 316 */ "Mili dis",
            /* 317 */ "Mili rank",
            /* 318 */ "Changed",
            /* 319 */ "Marriage registered",
            /* 320 */ "Marriage is not registered",
            /* 321 */ "Divorced",
            /* 322 */ "Birth",
            /* 323 */ "Adoption",
            /* 324 */ "Christening (child)",
            /* 325 */ "Graduation",
            /* 326 */ "Retirement",
            /* 327 */ "Naturalization",
            /* 328 */ "Emigration",
            /* 329 */ "Immigration",
            /* 330 */ "Census",
            /* 331 */ "Will",
            /* 332 */ "Probate of will",
            /* 333 */ "Death",
            /* 334 */ "Burial",
            /* 335 */ "Cremation",
            /* 336 */ "Fact",
            /* 337 */ "Physical description",
            /* 338 */ "National ID Number",
            /* 339 */ "Social security number",
            /* 340 */ "Number of children",
            /* 341 */ "Number of marriages",
            /* 342 */ "Property",
            /* 343 */ "Title",
            /* 344 */ "Travel",
            /* 345 */ "Hobby",
            /* 346 */ "Award",
            /* 347 */ "?",
            /* 348 */ "father",
            /* 349 */ "mother",
            /* 350 */ "husband",
            /* 351 */ "wife",
            /* 352 */ "son",
            /* 353 */ "daughter",
            /* 354 */ "grandfather",
            /* 355 */ "grandmother",
            /* 356 */ "grandson",
            /* 357 */ "granddaughter",
            /* 358 */ "brother",
            /* 359 */ "sister",
            /* 360 */ "son-in-law",
            /* 361 */ "daughter-in-law",
            /* 362 */ "father-in-law",
            /* 363 */ "mother-in-law",
            /* 364 */ "father-in-law",
            /* 365 */ "mother-in-law",
            /* 366 */ "uncle",
            /* 367 */ "aunt",
            /* 368 */ "nephew",
            /* 369 */ "niece",
            /* 370 */ "cousin",
            /* 371 */ "cousin",
            /* 372 */ "brother-in-law",
            /* 373 */ "sister-in-law",
            /* 374 */ "brother-in-law",
            /* 375 */ "sister-in-law",
            /* 376 */ "great",
            /* 377 */ "<?>",
            /* 378 */ "<?>",
            /* 379 */ "<?>",
            /* 380 */ "<?>",
            /* 381 */ "Image width is more than 65,000 pixels. Unable to save.",
            /* 382 */ "Back",
            /* 383 */ "Next",
            /* 384 */ "Previous",
            /* 385 */ "Next",
            /* 386 */ "You know that ...",
            /* 387 */ "Loading and searching places",
            /* 388 */ "(not selected)",
            /* 389 */ "Sample",
            /* 390 */ "For all people",
            /* 391 */ "Places of birth",
            /* 392 */ "Places of death",
            /* 393 */ "Places of residence",
            /* 394 */ "Only selected",
            /* 395 */ "Save image...",
            /* 396 */ "The format is not supported",
            /* 397 */ "Error loading data.",
            /* 398 */ "Print",
            /* 399 */ "Preview",
            /* 400 */ "Generation",
            /* 401 */ "contains",
            /* 402 */ "not contains",
            /* 403 */ "Exact",
            /* 404 */ "Before",
            /* 405 */ "After",
            /* 406 */ "Between",
            /* 407 */ "Period before",
            /* 408 */ "Period from",
            /* 409 */ "Period between",
            /* 410 */ "About",
            /* 411 */ "Calculated",
            /* 412 */ "Estimated",
            /* 413 */ "Engagement",
            /* 414 */ "Marriage",
            /* 415 */ "Public announcement of the marriage",
            /* 416 */ "Marriage contract",
            /* 417 */ "Obtaining permission to marry",
            /* 418 */ "Conclusion of the marriage agreement",
            /* 419 */ "Annulment of marriage",
            /* 420 */ "Filing of divorce",
            /* 421 */ "Divorce",
            /* 422 */ "Link to file",
            /* 423 */ "Saving to the archive",
            /* 424 */ "Saving to the store",
            /* 425 */ "-",
            /* 426 */ "Audio",
            /* 427 */ "Book",
            /* 428 */ "Card",
            /* 429 */ "Electronic",
            /* 430 */ "Fiche",
            /* 431 */ "Film",
            /* 432 */ "Magazine",
            /* 433 */ "Manuscript",
            /* 434 */ "Map",
            /* 435 */ "Newspaper",
            /* 436 */ "Photo",
            /* 437 */ "Tombstone",
            /* 438 */ "Video",
            /* 439 */ "- Unknown -",
            /* 440 */ "None",
            /* 441 */ "Low",
            /* 442 */ "Normal",
            /* 443 */ "High",
            /* 444 */ "Top",
            /* 445 */ "Defined",
            /* 446 */ "In progress",
            /* 447 */ "On hold",
            /* 448 */ "Problems",
            /* 449 */ "Completed",
            /* 450 */ "Withdrawn",
            /* 451 */ "Call",
            /* 452 */ "EMail",
            /* 453 */ "Fax",
            /* 454 */ "Letter",
            /* 455 */ "Tape",
            /* 456 */ "Visit",
            /* 457 */ "from",
            /* 458 */ "to",
            /* 459 */ "individual",
            /* 460 */ "family",
            /* 461 */ "source",
            /* 462 */ "other",
            /* 463 */ "Unreliable evidence",
            /* 464 */ "Questionable reliability of evidence",
            /* 465 */ "Circumstantial evidence",
            /* 466 */ "Direct and primary evidence",
            /* 467 */ "Research",
            /* 468 */ "Task",
            /* 469 */ "Correspondence",
            /* 470 */ "Correcting IDs",
            /* 471 */ "Checking format",
            /* 472 */ "Correction of identity records is required, continue?", // obsolete
            /* 473 */ "Number of objects in the master database: {0}",
            /* 474 */ "Synchronization is complete.",
            /* 475 */ "Identifying the patriarchs",
            /* 476 */ "Checking links",
            /* 477 */ "Archive not found, the data is not loaded",
            /* 478 */ "Scale",
            /* 479 */ "GEDKeeper GEDCOM files (*.ged)|*.ged|GEDKeeper GEDCOM encrypted files (*.geds)|*.geds",
            /* 480 */ "All files (*.*)|*.*",
            /* 481 */ "Images (*.bmp;*.gif;*.png;*.jpg)|*.bmp;*.gif;*.png;*.jpg",
            /* 482 */ "Scripts (*.lua)|*.lua",
            /* 483 */ "BMP files (*.bmp)|*.bmp|JPEG files (*.jpg)|*.jpg|EMF files (*.emf)|*.emf|PNG files (*.png)|*.png|GIF files (*.gif)|*.gif",
            /* 484 */ "Death causes",
            /* 485 */ "Pedigree",
            /* 486 */ "Parameter",
            /* 487 */ "Total",
            /* 488 */ "Men",
            /* 489 */ "Women",
            /* 490 */ "Fields filter",
            /* 491 */ "Year",
            /* 492 */ "PDF files (*.pdf)|*.pdf",
            /* 493 */ "RTF files (*.rtf)|*.rtf",
            /* 494 */ "HTML files (*.html)|*.html",
            /* 495 */ "Files backup",
            /* 496 */ "only previous file",
            /* 497 */ "Cutting off the branches",
            /* 498 */ "not",
            /* 499 */ "on years",
            /* 500 */ "in selected individuals",
            /* 501 */ "- all -",
            /* 502 */ "- not -",
            /* 503 */ "- any -",
            /* 504 */ "Master database",
            /* 505 */ "Database to merge",
            /* 506 */ "[current database]",
            /* 507 */ "There is a filter by a source. Do you want to add the source to a new individual?",
            /* 508 */ "There is a filter by a group. Do you want to add the group to a new individual?",
            /* 509 */ "Duplicates search",
            /* 510 */ "Invalid value",
            /* 511 */ "Operation",
            /* 512 */ "Compare databases",
            /* 513 */ "Merge databases",
            /* 514 */ "Split database",
            /* 515 */ "Merge records",
            /* 516 */ "Search unlinked namesakes",
            /* 517 */ "Check connection of families",
            /* 518 */ "Check database",
            /* 519 */ "Find the patriarchs",
            /* 520 */ "Manage places",
            /* 521 */ "Matching...",
            /* 522 */ "Check connection of families",
            /* 523 */ "all",
            /* 524 */ "living only ",
            /* 525 */ "dead only",
            /* 526 */ "Lived before",
            /* 527 */ "men only",
            /* 528 */ "women only",
            /* 529 */ "Mask of name",
            /* 530 */ "Mask of residence",
            /* 531 */ "Mask of the facts",
            /* 532 */ "Patriarchs only",
            /* 533 */ "Date is invalid",
            /* 534 */ "People",
            /* 535 */ "Years",
            /* 536 */ "Decade",
            /* 537 */ "Born",
            /* 538 */ "Died",
            /* 539 */ "Living",
            /* 540 */ "Dead",
            /* 541 */ "Average age",
            /* 542 */ "Life expectancy",
            /* 543 */ "Average number of children",
            /* 544 */ "Average age of the first child birth",
            /* 545 */ "Average number of marriages",
            /* 546 */ "Average age of marriage",
            /* 547 */ "Number of ancestors",
            /* 548 */ "Number of descendants",
            /* 549 */ "Number of generations of descendants",
            /* 550 */ "Birth years",
            /* 551 */ "Birth years (decades)",
            /* 552 */ "Death years",
            /* 553 */ "Death years (decades)",
            /* 554 */ "Distribution of the number of children",
            /* 555 */ "Age of first child birth",
            /* 556 */ "Age at marriage",
            /* 557 */ "Age difference of spouses",
            /* 558 */ "Select all links",
            /* 559 */ "Select family",
            /* 560 */ "Select ancestors",
            /* 561 */ "Select descendants",
            /* 562 */ "Merge",
            /* 563 */ "Search",
            /* 564 */ "Skip",
            /* 565 */ "Records",
            /* 566 */ "Search persons",
            /* 567 */ "Direct matching",
            /* 568 */ "Indistinct matching",
            /* 569 */ "Only by name / middle name (women only)",
            /* 570 */ "Check the year of birth",
            /* 571 */ "The accuracy of the name, %",
            /* 572 */ "Years inaccuracy",
            /* 573 */ "Repair",
            /* 574 */ "Generations of descendants, not less than",
            /* 575 */ "Set a sign",
            /* 576 */ "Register in list of locations",
            /* 577 */ "Similar surnames:",
            /* 578 */ "Similar names:",
            /* 579 */ "Selected personal records removed",
            /* 580 */ "Record",
            /* 581 */ "Problem",
            /* 582 */ "Solve",
            /* 583 */ "Perhaps died (age {0})",
            /* 584 */ "Sex is undefined",
            /* 585 */ "Year of birth more than year of death",
            /* 586 */ "The first marriage at the age of {0} years?",
            /* 587 */ "The first child was born at the age of {0} years?",
            /* 588 */ "Descendants",
            /* 589 */ "Generations",
            /* 590 */ "Number of links",
            /* 591 */ "Processing places",
            /* 592 */ "The place is already in the locations list",
            /* 593 */ "DNA markers",
            /* 594 */ "Average age of parents at the time first-born appeared, relative to parents age",
            /* 595 */ "Average age of parents at the time first-born appeared, relative to children age",
            /* 596 */ "Empty family record",
            /* 597 */ "backup each file revision",
            /* 598 */ "Loading",
            /* 599 */ "File saving error \"{0}\"{1}.",
            /* 600 */ "Background color...",
            /* 601 */ "Background image...",
            /* 602 */ "Persons in tree: {0}",
            /* 603 */ "View the error log",
            /* 604 */ "Ancestors Circle",
            /* 605 */ "Circle",
            /* 606 */ "Text color",
            /* 607 */ "Back color",
            /* 608 */ "Lines color",
            /* 609 */ "None",
            /* 610 */ "Locked",
            /* 611 */ "Confidential",
            /* 612 */ "Privacy",
            /* 613 */ "Send the error log",
            /* 614 */ "Charts",
            /* 615 */ "Fit window size",
            /* 616 */ "Zoom in",
            /* 617 */ "Zoom out",
            /* 618 */ "Ascertain a fact of death",
            /* 619 */ "Determine a sex",
            /* 620 */ "Plugins",
            /* 621 */ "Auto-trace relationships",
            /* 622 */ "No matches found",
            /* 623 */ "Find previous",
            /* 624 */ "Find next",
            /* 625 */ "The certainty index",
            /* 626 */ "Monthes of birth",
            /* 627 */ "Month",
            /* 628 */ "Reset",
            /* 629 */ "Field",
            /* 630 */ "Condition",
            /* 631 */ "Source is not specified",
            /* 632 */ "Slideshow",
            /* 633 */ "Previous record",
            /* 634 */ "Next record",
            /* 635 */ "Generation failed",
            /* 636 */ "Start",
            /* 637 */ "Stop",
            /* 638 */ "Demographics",
            /* 639 */ "Names",
            /* 640 */ "Name-nickname",
            /* 641 */ "Birth name",
            /* 642 */ "In immigration",
            /* 643 */ "Maiden name",
            /* 644 */ "Marriage name",
            /* 645 */ "Delete a name?",
            /* 646 */ "You cannot delete the single name",
            /* 647 */ "Quote",
            /* 648 */ "Place a bookmark on merged records",
            /* 649 */ "Show calendar near dates",
            /* 650 */ "Duplicate record",
            /* 651 */ "Attention: if you duplicate an individual, family links and associations are not copied",
            /* 652 */ "Show date type",
            /* 653 */ "Create a new tree file",
            /* 654 */ "Open the tree file",
            /* 655 */ "Save tree file",
            /* 656 */ "Add record (Ctrl+I)",
            /* 657 */ "Edit record (Ctrl+Enter)",
            /* 658 */ "Delete record (Ctrl+L)",
            /* 659 */ "Filtering list of records",
            /* 660 */ "Show a tree of ancestors",
            /* 661 */ "Show a tree of descendants",
            /* 662 */ "Show full tree",
            /* 663 */ "Pedigree",
            /* 664 */ "Pedigree by d'Aboville",
            /* 665 */ "Pedigree by Konovalov",
            /* 666 */ "Statistical analysis of data",
            /* 667 */ "Save the image",
            /* 668 */ "Display modes",
            /* 669 */ "Create a new script",
            /* 670 */ "Open a script",
            /* 671 */ "Save script",
            /* 672 */ "Execute the script",
            /* 673 */ "Attach the portrait",
            /* 674 */ "Detach the portrait",
            /* 675 */ "Join the family of parents",
            /* 676 */ "Edit the family of the parents",
            /* 677 */ "Disconnect family parents",
            /* 678 */ "Select or add the father",
            /* 679 */ "Disconnect father",
            /* 680 */ "Go to the record of the father",
            /* 681 */ "Select or add the mother",
            /* 682 */ "Disconnect mother",
            /* 683 */ "Go to the record of the mother",
            /* 684 */ "Copy name to clipboard",
            /* 685 */ "Select or add husband",
            /* 686 */ "Disconnect husband",
            /* 687 */ "Go to the record of the husband",
            /* 688 */ "Select or add wife",
            /* 689 */ "Disconnect wife",
            /* 690 */ "Go to the record of the wife",
            /* 691 */ "Choose an individual",
            /* 692 */ "Select or add a location",
            /* 693 */ "Remove the link to the place",
            /* 694 */ "Show the map",
            /* 695 */ "Select or add a record of source",
            /* 696 */ "Select the target record",
            /* 697 */ "If you use \"file\" or \"store\" storage type, you have to save new database file before continue",
            /* 698 */ "A file with the same name already exists",
            /* 699 */ "It's not a GEDSEC-compatible file",
            /* 700 */ "This individual has a few families. Children should be added via the spouses.",
            /* 701 */ "Export to Excel failed",
            /* 702 */ "Specific filter",
            /* 703 */ "Display lines",
            /* 704 */ "Type of comparison",
            /* 705 */ "Internal comparing (search for duplicates)",
            /* 706 */ "Comparing with another database",
            /* 707 */ "Analyze",
            /* 708 */ "Compare",
            /* 709 */ "Ignore dates",
            /* 710 */ "Chart",
            /* 711 */ "Family book",
            /* 712 */ "Table of contents",
            /* 713 */ "Records of individuals",
            /* 714 */ "Directories",
            /* 715 */ "Unable to auto-detect gender using name \"{0}\". Is it male gender?",
            /* 716 */ "Autosave every",
            /* 717 */ "minutes",
            /* 718 */ "Including number of generations",
            /* 719 */ "Hide empty segments",
            /* 720 */ "Descendants Circle",
            /* 721 */ "Choice of portrait",
            /* 722 */ "Relationship Calculator",
            /* 723 */ "{0} is {1} of {2}",
            /* 724 */ "Today is birthday of {0}",
            /* 725 */ "Tomorrow is birthday of {0}",
            /* 726 */ "Blood type",
            /* 727 */ "Hair color",
            /* 728 */ "Eyes color",
            /* 729 */ "mtDNA haplogroup",
            /* 730 */ "Y-DNA haplogroup",
            /* 731 */ "Record is locked by editor",
            /* 732 */ "Maiden surname",
            /* 733 */ "Married surname",
            /* 734 */ "Additional names",
            /* 735 */ "Extended women's surnames",
            /* 736 */ "Format \"Maiden (Married)\"",
            /* 737 */ "Format \"Married (Maiden)\"",
            /* 738 */ "Format \"Maiden\"",
            /* 739 */ "Format \"Married\"",
            /* 740 */ "H",
            /* 741 */ "W",
            /* 742 */ "Geocoder",
            /* 743 */ "Add father",
            /* 744 */ "Add mother",
            /* 745 */ "Show default portraits",
            /* 746 */ "A media file is added from the removable media. Continue?",
            /* 747 */ "Warn when adding media files from removable media",
            /* 748 */ "Load recently opened files at startup",
            /* 749 */ "Today is holiday \"{0}\"",
            /* 750 */ "Tomorrow is holiday \"{0}\"",
            /* 751 */ "{1} days remain before the holiday \"{0}\"",
            /* 752 */ "Holidays",
            /* 753 */ "Use embedded MediaPlayer",
            /* 754 */ "Allow direct references to media files",
            /* 755 */ "Copying file",
            /* 756 */ "Actions",
            /* 757 */ "Select and copy",
            /* 758 */ "№",
            /* 759 */ "Reports",
            /* 760 */ "Inverted tree",
            /* 761 */ "Dates of marriages",
            /* 762 */ "Automatically check for updates",
            /* 763 */ "Arc text",
            /* 764 */ "Album of Trees",
            /* 765 */ "Show places",
            /* 766 */ "Hide unknown spouses",
            /* 767 */ "Spacings",
            /* 768 */ "Margins",
            /* 769 */ "Branch distance",
            /* 770 */ "Generation distance",
            /* 771 */ "Spouse distance",
            /* 772 */ "Image size: {0} x {1} px",
            /* 773 */ "A data loop was detected: {0}",
            /* 774 */ "Father is listed as a child",
            /* 775 */ "Mother is listed as a child",
            /* 776 */ "There are duplicate children in the family",
            /* 777 */ "Select color",
            /* 778 */ "Autosort children",
            /* 779 */ "Autosort spouses",
            /* 780 */ "Check tree size allowed",
            /* 781 */ "Adopted",
            /* 782 */ "Birth",
            /* 783 */ "Foster",
            /* 784 */ "Detect Charset at load",
            /* 785 */ "Linkage type",
            /* 786 */ "Was born",
            /* 787 */ "Was born",
            /* 788 */ "GedML files (*.xml)|*.xml",
            /* 789 */ "Child detected without parents",
            /* 790 */ "Family record detected without family",
            /* 791 */ "Are you sure you want to remove multimedia file?",
            /* 792 */ "Storage not found",
            /* 793 */ "File \"{0}\" not found",
            /* 794 */ "Allow relative references to media files",
            /* 795 */ "Default media storage type",
            /* 796 */ "Allow delete media files from storage and archive",
            /* 797 */ "Allow delete media files by direct and relative references",
            /* 798 */ "Delete media files without confirm",
            /* 799 */ "Media record without files",
            /* 800 */ "Archive not found",

            /* 801 */ "RI:GeorgeKnight",
            /* 802 */ "USSR:WWII:Combatant",
            /* 803 */ "USSR:WWII:Killed in battle",
            /* 804 */ "USSR:WWII:Worker in rear",
            /* 805 */ "USSR:Repressed",
            /* 806 */ "Religion:Islam",
            /* 807 */ "Religion:Catholicism",
            /* 808 */ "Religion:Orthodoxy",
            /* 809 */ "Religion:The Old Believers",

            /* 810 */ "Relative link to file",
            /* 811 */ "Maximum number of backup revisions",
            /* 812 */ "First capital letter in names",
            /* 813 */ "Default depth",
            /* 814 */ "The dialog contains unsaved changes. Continue and save?",
            /* 815 */ "Warn for closing dialog with unsaved changes",
            /* 816 */ "Dotted lines of adopted children",
            /* 817 */ "Go to person record",
            /* 818 */ "Web address",
            /* 819 */ "Separate dates and places lines",
            /* 820 */ "Details",
            /* 821 */ "Copy XRef to clipboard",
            /* 822 */ "Bold font of names",

            /* 823 */ "Baptism",
            /* 824 */ "Bar mitzvah",
            /* 825 */ "Bat mitzvah",
            /* 826 */ "Blessing",
            /* 827 */ "Adult christening",
            /* 828 */ "Confirmation",
            /* 829 */ "First communion",
            /* 830 */ "Ordination",
            /* 831 */ "Restriction geo search by country",
            /* 832 */ "Separate depth",
            /* 833 */ "Ancestors",
            /* 834 */ "Matches not found",
            /* 835 */ "Short kinship form",
            /* 836 */ "Surname first in order",
            /* 837 */ "Borders",
            /* 838 */ "You've got version {0} of GEDKeeper. Would you like to update to the latest version {1}?",
            /* 839 */ "Providers",
            /* 840 */ "Coordinates",
            /* 841 */ "Load places",
            /* 842 */ "Only locality",
            /* 843 */ "Go to primary branch",
            /* 844 */ "Minimizing width",
            /* 845 */ "{0}. Continue?",
            /* 846 */ "Show age",
            /* 847 */ "Surname in capitals",
            /* 848 */ "|SVG files (*.svg)|*.svg",
            /* 849 */ "Certainty assessment algorithm",
            /* 850 */ "Localized calendar signatures",
            /* 851 */ "G|J|H|FR|R|I|?",
            /* 852 */ "Copy",
            /* 853 */ "Cut",
            /* 854 */ "Paste",
            /* 855 */ "Person `{0}` has a half link of spouse to family `{1}`",
            /* 856 */ "Person `{0}` has a half link of child to family `{1}`",
            /* 857 */ "Family `{0}` has a half link of husband to person `{1}`",
            /* 858 */ "Family `{0}` has a half link of wife to person `{1}`",
            /* 859 */ "Family `{0}` has a half link of child to parents family `{1}`",
            /* 860 */ "Family `{0}` has garbled spouses",
            /* 861 */ "USSR:CPSU member",
            /* 862 */ "Attention: when merging archive records, the name and address are not transferred! Proceed?",
            /* 863 */ "Use extra controls",
            /* 864 */ "Several families of parents `{0}`",
            /* 865 */ "Specials",
            /* 866 */ "Use inline images in SVG",
            /* 867 */ "Use extended notes",
            /* 868 */ "Keep rich names",
            /* 869 */ "Maximize chart windows",
            /* 870 */ "Extended tree",
            /* 871 */ "Fan chart (only ancestors)",
            /* 872 */ "Open in new window",
            /* 873 */ "Unknown place of person",
            /* 874 */ "Check person places",
            /* 875 */ "Highlight inaccessible files",
            /* 876 */ "The difference between spouses is abnormally high `{0}`",
            /* 877 */ "The difference between siblings is abnormally high `{0}`",

            /* 878 */ "Find and Replace",
            /* 879 */ "Find",
            /* 880 */ "Replace with",
            /* 881 */ "Match case",
            /* 882 */ "Match wildcards",
            /* 883 */ "Whole word",
            /* 884 */ "Replace",
            /* 885 */ "Replace all",
            /* 886 */ "Property",

            /* 887 */ "Themes",
            /* 888 */ "Tree note",
            /* 889 */ "Show special notes",
            /* 890 */ "No goal record specified",
            /* 891 */ "Invalid file name",
            /* 892 */ "Search and filter by all names",
            /* 893 */ "Data Quality",
            /* 894 */ "Fragment {0}, size {1}, quality {2}",
            /* 895 */ "Refresh",
            /* 896 */ "Reset filter",
            /* 897 */ "Shorten date ranges",
            /* 898 */ "Press Enter to save the filter",

            /* 899 */ "Today is jubilee of {0}",
            /* 900 */ "Tomorrow is jubilee of {0}",
            /* 901 */ "{1} days remain before the jubilee of {0}",
        };

        private static readonly LangManager fLangMan = new LangManager();

        public static string LS(LSID lsid)
        {
            return fLangMan.LS(lsid);
        }

        public static string LS(LSID lsid, params object[] args)
        {
            return string.Format(fLangMan.LS(lsid), args);
        }

        public static void DefInit()
        {
            fLangMan.DefInit(LSDefList);
        }

        public static bool LoadFromFile(string fileName)
        {
            return fLangMan.LoadFromFile(fileName, 1);
        }

        public static void SaveDefaultLanguage()
        {
            StreamWriter lf = new StreamWriter(GKUtils.GetLangsPath() + "english.sample2", false, Encoding.UTF8);
            try
            {
                lf.WriteLine(";" + LS_DEF_CODE.ToString() + "," + LS_DEF_SIGN + "," + LS_DEF_NAME);
                for (LSID i = LSID.LSID_First; i <= LSID.LSID_Last; i++)
                {
                    string ls = LSDefList[(int)i - 1];
                    if (!string.IsNullOrEmpty(ls)) {
                        lf.WriteLine(ConvertHelper.AdjustNumber((int)i, 3) + "=" + ls);
                    }
                }
            }
            finally
            {
                lf.Close();
            }
        }
    }

    public class LangManager : ILangMan
    {
        private readonly Dictionary<int, string> fList;

        public LangManager()
        {
            fList = new Dictionary<int, string>();
        }

        public string LS(Enum lsid)
        {
            int idx = ((IConvertible)lsid).ToInt32(null);
            string res;
            return (fList.TryGetValue(idx, out res)) ? res : "?";
        }

        public bool LoadFromFile(string fileName, int offset = 0)
        {
            bool result = false;

            if (File.Exists(fileName))
            {
                fList.Clear();

                using (StreamReader lngFile = new StreamReader(fileName, Encoding.UTF8))
                {
                    bool xt = false;

                    string st = lngFile.ReadLine();
                    if (!string.IsNullOrEmpty(st) && st[0] == ';')
                    {
                        st = st.Remove(0, 1);
                        string[] lngParams = st.Split(',');
                        if (lngParams.Length < 3)
                            throw new GKException("Header is incorrect");

                        xt = (lngParams.Length == 4 && lngParams[3] == "xt");
                    }

                    int i = 0 + (offset);
                    while (lngFile.Peek() != -1)
                    {
                        st = lngFile.ReadLine().Trim();

                        if (xt) {
                            // allowed empty and comment strings
                            if (string.IsNullOrEmpty(st) || st[0] == ';') {
                                continue;
                            }

                            string[] parts = st.Split('=');
                            i = int.Parse(parts[0]);
                            fList.Add(i, parts[1]);
                        } else {
                            fList.Add(i, st);
                            i++;
                        }
                    }
                    result = true;
                }
            }

            return result;
        }

        internal void DefInit(string[] source)
        {
            fList.Clear();

            for (LSID id = LSID.LSID_First; id <= LSID.LSID_Last; id++)
            {
                int idx = (int)id;
                fList.Add(idx, source[idx - 1]);
            }
        }
    }
}
